@@ -2,8 +2,6 @@
 set -euo pipefail
 
 # Determine configuration
-export RUST_TEST_NOCAPTURE=1
-export RUST_BACKTRACE=1
 export RUSTFLAGS="-D warnings"
 export CARGO_INCREMENTAL=0
 export CARGO_EXTRA_FLAGS="--all-features"
@@ -24,8 +22,11 @@ function run_tests {
 
   ./miri test --locked
   if [ -z "${MIRI_TEST_TARGET+exists}" ]; then
-    # Only for host architecture: tests with MIR optimizations
-    MIRIFLAGS="-Z mir-opt-level=3" ./miri test --locked
+    # Only for host architecture: tests with optimizations (`-O` is what cargo passes, but crank MIR
+    # optimizations up all the way).
+    # Optimizations change diagnostics (mostly backtraces), so we don't check them
+    #FIXME(#2155): we want to only run the pass and panic tests here, not the fail tests.
+    MIRIFLAGS="-O -Zmir-opt-level=4" MIRI_SKIP_UI_CHECKS=1 ./miri test --locked -- tests/{pass,panic}
   fi
 
   # On Windows, there is always "python", not "python3" or "python2".
@@ -56,7 +57,6 @@ case $HOST_TARGET in
     ;;
   i686-pc-windows-msvc)
     MIRI_TEST_TARGET=x86_64-unknown-linux-gnu run_tests
-    MIRI_TEST_TARGET=x86_64-apple-darwin run_tests
     ;;
   *)
     echo "FATAL: unknown OS"
